package org.apache.ojb.broker;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.time.StopWatch;
import org.apache.commons.lang.ClassUtils;
import org.apache.ojb.broker.metadata.ClassDescriptor;
import org.apache.ojb.broker.metadata.CollectionDescriptor;
import org.apache.ojb.broker.metadata.ObjectReferenceDescriptor;
import org.apache.ojb.broker.query.Criteria;
import org.apache.ojb.broker.query.Query;
import org.apache.ojb.broker.query.QueryFactory;
import org.apache.ojb.junit.PBTestCase;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

/**
 * Test (non-decomposed) M:N relations.
 *
 * IMPORTANT NOTE: The global runtime metadata changes made by this test case
 * are NOT recommended in multithreaded environments, because they are global
 * and each thread will be affected.
 *
 * @author <a href="mailto:arminw@apache.org">Armin Waibel</a>
 * @version $Id: M2NTest.java,v 1.7.2.4 2004/11/27 02:17:47 arminw Exp $
 */
public class M2NTest extends PBTestCase
{
    static final int NONE = ObjectReferenceDescriptor.CASCADE_NONE;
    static final int LINK = ObjectReferenceDescriptor.CASCADE_LINK;
    static final int OBJECT = ObjectReferenceDescriptor.CASCADE_OBJECT;

    int actorCount = 200;
    int movieCount = 100;

    public static void main(String[] args)
    {
        junit.textui.TestRunner.main(new String[]{M2NTest.class.getName()});
    }

    public void tearDown() throws Exception
    {
        if(broker != null)
        {
            changeActorCollectionDescriptorTo(false, NONE, NONE, false);
            changeMovieCollectionDescriptorTo(false, NONE, NONE, false);
            broker.close();
        }
        super.tearDown();
    }

    public void testAutoUpdateDeleteSettings()
    {
        changeActorCollectionDescriptorTo(false, false, false, false);
        CollectionDescriptor ord = broker.getClassDescriptor(Actor.class)
                .getCollectionDescriptorByName("movies");
        assertEquals(CollectionDescriptor.CASCADE_LINK, ord.getCascadingStore());
        assertEquals(CollectionDescriptor.CASCADE_LINK, ord.getCascadingDelete());
        assertEquals(false, ord.getCascadeStore());
        assertEquals(false, ord.getCascadeDelete());

        changeActorCollectionDescriptorTo(false, true, true, false);
        ord = broker.getClassDescriptor(Actor.class)
                .getCollectionDescriptorByName("movies");
        assertEquals(CollectionDescriptor.CASCADE_OBJECT, ord.getCascadingStore());
        assertEquals(CollectionDescriptor.CASCADE_OBJECT, ord.getCascadingDelete());
        assertEquals(true, ord.getCascadeStore());
        assertEquals(true, ord.getCascadeDelete());
    }

    public void testMassStoreUpdateAutomatic()
    {

        long testPeriod = 0;

        String postfix = "testMassStoreUpdateAutomatic_" + System.currentTimeMillis();
        changeMovieCollectionDescriptorTo(true, OBJECT, OBJECT, false);
        changeActorCollectionDescriptorTo(true, OBJECT, OBJECT, false);

        Movie movie = buildMovieWithActors(postfix, actorCount);
        Actor actor = buildActorWithMovies(postfix, movieCount);

        List actors = new ArrayList(movie.getActors());
        actors.add(actor);
        movie.setActors(actors);

        MovieManageableCollection movies = actor.getMovies();
        movies.add(movie);

        StopWatch watch = new StopWatch();
        watch.start();
        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateAutomatic] Time to store "+(actorCount + movieCount)+" m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();

        watch.reset();
        watch.start();
        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(movieCount + 1, resultMovie.size());
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateAutomatic] Time to query "+movieCount+" m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();

        watch.reset();
        watch.start();
        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(actorCount + 1, resultActor.size());
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateAutomatic] Time to query "+actorCount+" m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();

        Query queryRole = roleQueryActorOrMovieMatch(actor, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(actorCount + movieCount + 1, resultRole.size());

        //*****************************
        // update movie
        movie.setActors(new ArrayList());
        watch.reset();
        watch.start();
        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateAutomatic] Time to update object with "+actorCount + " m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();
        //*****************************

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        movie = (Movie) broker.getObjectByIdentity(oid);

        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(movieCount + 1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(actorCount + 1, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(movieCount, resultRole.size());

        assertNotNull(movie);
        assertEquals(0, movie.getActors().size());

        //*****************************
        // remove actor
        movie.setActors(new ArrayList());
        watch.reset();
        watch.start();
        broker.beginTransaction();
        broker.delete(actor);
        broker.commitTransaction();
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateAutomatic] Time to remove object with "+ movieCount + " m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();
        //*****************************

        broker.clearCache();
        oid = new Identity(actor, broker);
        actor = (Actor) broker.getObjectByIdentity(oid);

        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());

        // we never delete these actor objects
        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(actorCount, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());

        assertNull(actor);

        broker.beginTransaction();
        broker.delete(movie);
        broker.commitTransaction();

        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateAutomatic] Total test time is "+ testPeriod+" ms");
        System.out.println("");
    }


    public void testMassStoreUpdateLinking()
    {
        long testPeriod = 0;

        String postfix = "testMassStoreUpdateLinking" + System.currentTimeMillis();
        changeMovieCollectionDescriptorTo(true, NONE, OBJECT, false);
        changeActorCollectionDescriptorTo(true, NONE, OBJECT, false);

        Movie movie = buildMovieWithActors(postfix, actorCount);
        Actor actor = buildActorWithMovies(postfix, movieCount);

        List actors = new ArrayList(movie.getActors());
        actors.add(actor);
        movie.setActors(actors);

        MovieManageableCollection movies = actor.getMovies();
        movies.add(movie);

        StopWatch watch = new StopWatch();
        watch.start();
        broker.beginTransaction();
        broker.store(movie);
        for(int i = 0; i < actors.size(); i++)
        {
            broker.store(actors.get(i));
        }
        MovieManageableCollection actorMovies = actor.getMovies();
        for(int i = 0; i < actorMovies.size(); i++)
        {
             broker.store(actorMovies.get(i));
        }
        broker.serviceBrokerHelper().link(movie, true);
        broker.serviceBrokerHelper().link(actor, true);
        broker.commitTransaction();
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateLinking] Time to store "+(actorCount + movieCount)+" m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();

        watch.reset();
        watch.start();
        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(movieCount + 1, resultMovie.size());
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateLinking] Time to query "+movieCount+" m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();

        watch.reset();
        watch.start();
        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(actorCount + 1, resultActor.size());
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateLinking] Time to query "+actorCount+" m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();

        Query queryRole = roleQueryActorOrMovieMatch(actor, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(actorCount + movieCount + 1, resultRole.size());

        //*****************************
        // update movie
        movie.setActors(new ArrayList());
        watch.reset();
        watch.start();
        broker.beginTransaction();
        broker.serviceBrokerHelper().unlink(movie);
        broker.store(movie);
        broker.serviceBrokerHelper().link(movie, false);
        broker.commitTransaction();
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateLinking] Time to update object with "+actorCount + " m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();
        //*****************************

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        movie = (Movie) broker.getObjectByIdentity(oid);

        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(movieCount + 1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(actorCount + 1, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(movieCount, resultRole.size());

        assertNotNull(movie);
        assertEquals(0, movie.getActors().size());

        //*****************************
        // remove actor
        movie.setActors(new ArrayList());
        watch.reset();
        watch.start();
        broker.beginTransaction();
        broker.delete(actor);
        broker.commitTransaction();
        watch.stop();
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateLinking] Time to remove object with "+ movieCount + " m:n objects=" + watch.getTime());
        testPeriod += watch.getTime();
        //*****************************

        broker.clearCache();
        oid = new Identity(actor, broker);
        actor = (Actor) broker.getObjectByIdentity(oid);

        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());

        // we never delete these actor objects
        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(actorCount, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());

        assertNull(actor);

        broker.beginTransaction();
        broker.delete(movie);
        broker.commitTransaction();

        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());
        System.out.println("[" + ClassUtils.getShortClassName(this.getClass())
                + "#testMassStoreUpdateLinking] Total test time is "+ testPeriod+" ms");
        System.out.println("");
    }

    //============================================================
    // auto-retrieve true / auto-update/auto-delete LINK
    //============================================================
    // test needs refactoring, LINK settings doesn't allow to store
    // an object graph
    public void XXXtestStoreAddUpdateDeleteTLLF()
    {
        changeMovieCollectionDescriptorTo(true, LINK, LINK, false);
        changeActorCollectionDescriptorTo(true, LINK, LINK, false);
        String postfix = "_testStoreTLLF_" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);
        doTestStoreAddUpdateDeleteTLLX(movie, postfix);
    }

    public void XXXtestStoreAddUpdateDeleteTLLT()
    {
        changeMovieCollectionDescriptorTo(true, LINK, LINK, true);
        changeActorCollectionDescriptorTo(true, LINK, LINK, false);
        String postfix = "_testStoreTLLF_" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);
        doTestStoreAddUpdateDeleteTLLX(movie, postfix);
    }


    public void doTestStoreAddUpdateDeleteTLLX(Movie movie, String postfix)
    {
        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();
        Identity movieOID = broker.serviceIdentity().buildIdentity(movie);

        /*
        auto-update setting is LINK, but it behaves same way as OBJECT
        */
        broker.clearCache();
        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(3, resultMovie.size());

        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        // first created actor
        Actor actor = (Actor) movie.getActors().iterator().next();
        Query queryRole = queryRole(actor, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(5, resultRole.size());

        broker.clearCache();
        broker.beginTransaction();
        Movie newMovie = (Movie) broker.getObjectByIdentity(movieOID);
        newMovie.setTitle("updated_title_" + postfix);
        Iterator it = newMovie.getActors().iterator();
        while(it.hasNext())
        {
            Actor a = (Actor) it.next();
            a.setName("updated_name_" + postfix);
        }
        ArrayList list = new ArrayList(newMovie.getActors());
        list.add(new Actor("updated_name_" + postfix));
        newMovie.setActors(list);
        broker.store(newMovie);
        broker.commitTransaction();

        broker.clearCache();
        queryMovie = queryMovie(postfix);
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(3, resultMovie.size());

        queryActor = queryActor(postfix);
        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(4, resultActor.size());

        // first created actor
        actor = (Actor) movie.getActors().iterator().next();
        queryRole = queryRole(actor, movie);
        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(6, resultRole.size());

        newMovie = (Movie) broker.getObjectByIdentity(movieOID);
        assertEquals("updated_title_"+postfix, newMovie.getTitle());
        it = newMovie.getActors().iterator();
        while(it.hasNext())
        {
            Actor a = (Actor) it.next();
            assertEquals("updated_name_" + postfix, a.getName());
        }

        // now we delete the movie object
        broker.beginTransaction();
        broker.delete(newMovie);
        broker.commitTransaction();

        broker.clearCache();
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(2, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(4, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(2, resultRole.size());
    }

    //============================================================
    // auto-update / auto-retrieve false
    //============================================================

    /**
     * All auto-xxx settings are false, thus expect that all thing must be
     * done by hand
     */
    public void testStoreFFFF()
    {
        changeMovieCollectionDescriptorTo(false, NONE, NONE, false);
        changeActorCollectionDescriptorTo(false, NONE, NONE, false);
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActors(postfix);
        doTestStoreFFFX(movie, postfix);
    }

    /**
     * All auto-xxx settings are false, thus expect that all thing must be
     * done by hand, proxy true should not affect anything
     */
    public void testStoreFFFT()
    {
        changeMovieCollectionDescriptorTo(false, NONE, NONE, true);
        changeActorCollectionDescriptorTo(false, NONE, NONE, false);
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActors(postfix);
        doTestStoreFFFX(movie, postfix);
    }

    /**
     * All auto-xxx settings are false, thus expect that all thing must be
     * done by hand. Actors has a collection of movies too - should be ignored
     * in this case.
     */
    public void testStoreFFFF_2()
    {
        changeMovieCollectionDescriptorTo(false, NONE, NONE, false);
        changeActorCollectionDescriptorTo(false, NONE, NONE, false);
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);
        doTestStoreFFFX(movie, postfix);
    }

    /**
     * All auto-xxx settings are false, thus expect that all thing must be
     * done by hand, proxy true should not affect anything. Actors has a
     * collection of movies too - should be ignored in this case.
     */
    public void testStoreFFFT_2()
    {
        changeMovieCollectionDescriptorTo(false, NONE, NONE, true);
        changeActorCollectionDescriptorTo(false, NONE, NONE, false);
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);
        doTestStoreFFFX(movie, postfix);
    }

    public void doTestStoreFFFX(Movie movie, String postfix)
    {
        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        /*
        all auto-xxx settings are false, so only the movie object should be
        stored - no other inserts!
        */
        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        // auto-update is false, thus we don't expect an Actor object
        assertEquals(0, resultActor.size());

        Query queryRole = queryRole(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        // auto-update is false, thus we don't expect Role objects
        assertEquals(0, resultRole.size());

        broker.beginTransaction();
        /*
        now we store the right-side objects and the intermediary entries
        */
        Iterator it = movie.getActors().iterator();
        while(it.hasNext())
        {
            Object actor = it.next();
            broker.store(actor);
        }
        // now both side exist and we can link the references
        broker.serviceBrokerHelper().link(movie, "actors", true);
        /*
        alternative call
        broker.serviceBrokerHelper().link(movie, true);
        */
        broker.commitTransaction();

        /*
        now we expect all stored objects
        */
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        Movie readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        // auto-retrieve false
        assertTrue(readMovie.getActors() == null || readMovie.getActors().size() == 0);

        broker.retrieveAllReferences(readMovie);
        assertEquals(3, readMovie.getActors().size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());
        Actor readActor = (Actor) resultActor.iterator().next();
        // auto-retrieve false
        assertTrue(readActor.getMovies() == null || readActor.getMovies().size() == 0);
        broker.retrieveAllReferences(readActor);
        assertEquals(1, readActor.getMovies().size());

        // We try to delete all objects
        // first do unlink the m:n references
        broker.beginTransaction();
        broker.serviceBrokerHelper().unlink(readMovie, "actors");
        /*
        alternative call
        broker.serviceBrokerHelper().unlink(readMovie);
        */
        broker.commitTransaction();

        broker.clearCache();
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());

        // now we delete the n- and m-side objects
        broker.beginTransaction();
        Iterator iter = movie.getActors().iterator();
        while(iter.hasNext())
        {
            broker.delete(iter.next());
        }
        broker.delete(movie);
        broker.commitTransaction();

        broker.clearCache();
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(0, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());
    }

    //============================================================
    // auto-retrieve true / auto-update/auto-delete false
    //============================================================
    /**
     * auto-retrieve is true, other false. proxy true/false should not affect
     * anything.
     */
    public void testStoreTFFF()
    {
        changeMovieCollectionDescriptorTo(true, NONE, NONE, false);
        changeActorCollectionDescriptorTo(true, NONE, NONE, false);
        doTestStoreTFFX();
    }

    /**
     * auto-retrieve is true, other false. proxy true/false should not affect
     * anything.
     */
    public void testStoreTFFT()
    {
        changeMovieCollectionDescriptorTo(true, NONE, NONE, true);
        changeActorCollectionDescriptorTo(true, NONE, NONE, false);
        doTestStoreTFFX();
    }

    public void doTestStoreTFFX()
    {
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);

        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        /*
        auto-update settings is false, so only the movie object should be
        stored - no other inserts!
        */
        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        // auto-update is false, thus we don't expect an Actor object
        assertEquals(0, resultActor.size());

        Query queryRole = queryRole(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        // auto-update is false, thus we don't expect Role objects
        assertEquals(0, resultRole.size());

        broker.beginTransaction();
        /*
        now we store the right-side objects and the intermediary entries
        */
        Iterator it = movie.getActors().iterator();
        while(it.hasNext())
        {
            Object actor = it.next();
            broker.store(actor);
        }
        // now both side exist and we can link the references
        broker.serviceBrokerHelper().link(movie, "actors", true);
        /*
        alternative call
        broker.serviceBrokerHelper().link(movie, true);
        */
        broker.commitTransaction();

        /*
        now we expect all stored objects
        */
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        Movie readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        // auto-retrieve true
        assertTrue(readMovie.getActors() != null);
        assertEquals(3, readMovie.getActors().size());

        // We try to delete all objects
        // first do unlink the m:n references
        broker.beginTransaction();
        broker.serviceBrokerHelper().unlink(readMovie, "actors");
        /*
        alternative call
        broker.serviceBrokerHelper().unlink(readMovie);
        */
        broker.commitTransaction();

        broker.clearCache();
        // TODO: replace this with query below (when prefetching bug was solved)
        Movie movieLookup = (Movie) broker.getObjectByIdentity(new Identity(movie, broker));
        assertNotNull(movieLookup);
        //resultMovie = broker.getCollectionByQuery(queryMovie);
        //assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());

        // now we delete the n- and m-side objects
        broker.beginTransaction();
        Iterator iter = movie.getActors().iterator();
        while(iter.hasNext())
        {
            broker.delete(iter.next());
        }
        broker.delete(movie);
        broker.commitTransaction();

        broker.clearCache();
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(0, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());
    }




    //============================================================
    // auto-update / auto-retrieve true
    //============================================================

    public void testStoreTTFF()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, false);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreTTXX();
    }

    public void testStoreTTFT()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, true);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreTTXX();
    }

    public void doTestStoreTTXX()
    {
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActors(postfix);

        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        Query queryRole = queryRole(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        Movie readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        assertEquals(3, readMovie.getActors().size());
    }

    /**
     * movies with back-references
     * auto-update = OBJECT 
     */
    public void testStoreTTFT_2()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, true);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreTTXX_2();
    }

    public void doTestStoreTTXX_2()
    {
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);

        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(3, resultMovie.size());

        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        Query queryRole = queryRole(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        Movie readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        assertEquals(3, readMovie.getActors().size());
    }
    
    public void testStoreUpdateTTFF()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, false);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreUpdateTTXX();
    }

    public void testStoreUpdateTTFF_2()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, true);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreUpdateTTXX();
    }

    public void doTestStoreUpdateTTXX()
    {
        String postfix = "doTestStoreUpdateTTXX_" + System.currentTimeMillis();
        Movie movie = buildMovieWithActors(postfix);

        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        Query queryRole = queryRole(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        //*****************************
        // remove all actors
        movie.setActors(new ArrayList());
        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();
        //*****************************

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        movie = (Movie) broker.getObjectByIdentity(oid);

        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());

        assertNotNull(movie);
        assertEquals(0, movie.getActors().size());
    }

    public void testStoreUpdateActorTTFF()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, false);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);

        String postfix = "" + System.currentTimeMillis();
        Actor actor = buildActorWithMovies(postfix);

        broker.beginTransaction();
        broker.store(actor);
        broker.commitTransaction();

        broker.clearCache();
        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(3, resultMovie.size());

        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(1, resultActor.size());

        Query queryRole = queryRole(actor, null);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(actor, broker);
        Actor loadedActor = (Actor) broker.getObjectByIdentity(oid);
        assertNotNull(loadedActor);
        MovieManageableCollection col = loadedActor.getMovies();
        assertNotNull(col);
        Movie movie = col.get(0);
    }

    public void testAddNewEntriesTTTF()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, OBJECT, false);
        changeActorCollectionDescriptorTo(true, OBJECT, OBJECT, false);
        doTestStoreUpdateTTXX();
    }

    public void testAddNewEntriesTTTT()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, OBJECT, true);
        // default proxy does not work for user defined collection
        changeActorCollectionDescriptorTo(true, OBJECT, OBJECT, false);
        doTestStoreUpdateTTXX();
    }

    public void doTestAddNewEntries()
    {
        String postfix = "" + System.currentTimeMillis();

        /*
        Returns 1 movie object with 3 actor objects and one actor object with
        back-reference to movie object + 2 new movies
        */
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);
        Actor a_1 = new Actor("testAddNewEntries_"+postfix);
        Actor a_2 = new Actor("testAddNewEntries_"+postfix);
        Actor a_3 = new Actor("testAddNewEntries_"+postfix);
        /*
        all in all we expect 3 movie, 6 actor, 3 role entries after first
        store.
        */

        broker.beginTransaction();
        broker.store(movie);
        broker.store(a_1);
        broker.store(a_2);
        broker.store(a_3);
        broker.commitTransaction();

        broker.clearCache();

        Query queryMovie = queryMovie(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(3, resultMovie.size());

        Query queryActor = queryActor(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(6, resultActor.size());

        Query queryRole = queryRole(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        Movie readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        assertEquals(3, readMovie.getActors().size());

        /*
        we add 2 existing actor an movie object, thus we expect
        3 movie, 6 actor, 5 role entries after store.
        And next lookup of movie we expect 5 dependend actor objects
        */
        movie.getActors().add(a_1);
        movie.getActors().add(a_2);
        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        broker.clearCache();

        queryMovie = queryMovie(postfix);
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(3, resultMovie.size());

        queryActor = queryActor(postfix);
        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(6, resultActor.size());

        queryRole = queryRole(null, movie);
        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(5, resultRole.size());

        broker.clearCache();
        oid = new Identity(movie, broker);
        readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        assertEquals(5, readMovie.getActors().size());

        /*
        on delete we expect that all entries are deleted except the single
        actor which have no references to any movie object
        */
        broker.beginTransaction();
        broker.delete(movie);
        broker.commitTransaction();

        broker.clearCache();
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(1, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());
    }

    //=======================================================================
    // helper methods
    //=======================================================================
    Query queryMovie(String postfix)
    {
        Criteria c = new Criteria();
        c.addLike("idStr", "%" + postfix + "%");
        Query q = QueryFactory.newQuery(Movie.class, c);
        return q;
    }

    Query queryActor(String postfix)
    {
        Criteria c = new Criteria();
        c.addLike("name", "%" + postfix + "%");
        Query q = QueryFactory.newQuery(Actor.class, c);
        return q;
    }

    Query queryRole(Actor actor, Movie movie)
    {
        Criteria c = new Criteria();
        if(actor != null) c.addEqualTo("actorId", actor.getId());
        if(movie != null && actor != null)
        {
            Criteria c2 = new Criteria();
            c2.addEqualTo("movieIntId", movie.getIdInt());
            c2.addEqualTo("movieStrId", movie.getIdStr());
            c.addOrCriteria(c2);
        }
        else if(movie != null)
        {
            c.addEqualTo("movieIntId", movie.getIdInt());
            c.addEqualTo("movieStrId", movie.getIdStr());
        }
        Query q = QueryFactory.newQuery(Role.class, c);
        return q;
    }

    Query roleQueryActorOrMovieMatch(Actor actor, Movie movie)
    {
        Criteria c_1 = new Criteria();
        Criteria c_2 = new Criteria();
        if(actor != null) c_1.addEqualTo("actorId", actor.getId());
        if(movie != null)
        {
            c_2.addEqualTo("movieIntId", movie.getIdInt());
            c_2.addEqualTo("movieStrId", movie.getIdStr());
        }
        if(actor != null)
        {
            c_2.addOrCriteria(c_1);
        }
        else
        {
            c_2 = c_1;
        }
        Query q = QueryFactory.newQuery(Role.class, c_2);
        return q;
    }

    /**
     * Returns 1 movie object with 3 actor objects
     */
    Movie buildMovieWithActors(String postfixId)
    {
        Movie m = new MovieImpl(postfixId,
                "Dr. Strangelove or: How I Learned to Stop Worrying and Love the Bomb " + postfixId,
                "An insane general starts a process to nuclear holocaust that a war" +
                " room of politicians and generals frantically try to stop. " + postfixId);
        Actor a1 = new Actor("Peter Sellers " + postfixId);
        Actor a2 = new Actor("George C. Scott " + postfixId);
        Actor a3 = new Actor("Sterling Hayden " + postfixId);
        ArrayList list = new ArrayList();
        list.add(a1);
        list.add(a2);
        list.add(a3);
        m.setActors(list);
        return m;
    }

    /**
     * Returns 1 movie object with 3 actor objects
     */
    Movie buildMovieWithActors(String postfixId, int actorCount)
    {
        Movie m = new MovieImpl(postfixId, "Movie with "+ actorCount+" actors_" + postfixId, "none");

        ArrayList list = new ArrayList();
        for(int i = 0; i < actorCount; i++)
        {
            Actor a = new Actor("A bad actor_" + postfixId);
            list.add(a);
        }
        m.setActors(list);
        return m;
    }

    /**
     * Returns 1 movie object with 3 actor objects
     */
    Actor buildActorWithMovies(String postfixId, int movieCount)
    {
        Actor a = new Actor(postfixId+"_Actor play in "+ movieCount+" movies");

        MovieManageableCollection list = new MovieManageableCollection();
        for(int i = 0; i < movieCount; i++)
        {
            Movie m  = new MovieImpl(postfixId, "A bad movie_" + postfixId, "none");
            list.add(m);
        }
        a.setMovies(list);
        return a;
    }

    /**
     * Returns 1 movie object with 3 actor objects and one actor object with
     * back-reference to movie object + 2 new movies
     */
    Movie buildMovieWithActorsAndBackReferences(String postfixId)
    {
        Movie m = new MovieImpl(postfixId,
                "Dr. Strangelove or: How I Learned to Stop Worrying and Love the Bomb " + postfixId,
                "An insane general starts a process to nuclear holocaust that a war" +
                " room of politicians and generals frantically try to stop. " + postfixId);
        Actor a1 = new Actor("Peter Sellers " + postfixId);
        Actor a2 = new Actor("George C. Scott " + postfixId);
        Actor a3 = new Actor("Sterling Hayden " + postfixId);
        ArrayList list = new ArrayList();
        list.add(a1);
        list.add(a2);
        list.add(a3);
        m.setActors(list);

        Movie m1 = new MovieImpl(postfixId + "", "A Shot in the Dark",
                "As murder follows murder, beautiful Maria is the obvious suspect...");
        Movie m2 = new MovieImpl(postfixId + "", "The Pink Panther",
                " In the first movie starring Peter Sellers as the bumbling Inspector Clouseau...");

        MovieManageableCollection mlist1 = new MovieManageableCollection();
        mlist1.add(m);
        mlist1.add(m1);
        mlist1.add(m2);
        MovieManageableCollection mlist2 = new MovieManageableCollection();
        MovieManageableCollection mlist3 = new MovieManageableCollection();
        a1.setMovies(mlist1);
        a2.setMovies(mlist2);
        a3.setMovies(mlist3);

        return m;
    }

    Actor buildActorWithMovies(String postfixId)
    {
        Actor a = new Actor("John Cusack" + postfixId);
        MovieManageableCollection list = new MovieManageableCollection();
        list.add(new MovieImpl("a_" + postfixId, "High Fidelity", "A comedy about fear of commitment, hating your job..." + postfixId));
        list.add(new MovieImpl("b_" + postfixId, "Identity", "When a nasty storm hits a hotel, ten strangers are stranded within ..." + postfixId));
        list.add(new MovieImpl("c_" + postfixId, "Grosse Pointe Blank", "Martin Blank is a professional assassin. He is sent on a mission to a small Detroit ..." + postfixId));
        a.setMovies(list);
        return a;
    }

    void changeActorCollectionDescriptorTo(boolean autoRetrieve, int autoUpdate, int autoDelete, boolean proxy)
    {
        ClassDescriptor cld = broker.getClassDescriptor(Actor.class);
        CollectionDescriptor cod = (CollectionDescriptor) cld.getCollectionDescriptors().get(0);
        cod.setLazy(proxy);
        cod.setCascadeRetrieve(autoRetrieve);
        cod.setCascadingStore(autoUpdate);
        cod.setCascadingDelete(autoDelete);
    }

    void changeActorCollectionDescriptorTo(boolean autoRetrieve, boolean autoUpdate, boolean autoDelete, boolean proxy)
    {
        ClassDescriptor cld = broker.getClassDescriptor(Actor.class);
        CollectionDescriptor cod = (CollectionDescriptor) cld.getCollectionDescriptors().get(0);
        cod.setLazy(proxy);
        cod.setCascadeRetrieve(autoRetrieve);
        cod.setCascadeStore(autoUpdate);
        cod.setCascadeDelete(autoDelete);
    }

    void changeMovieCollectionDescriptorTo(boolean autoRetrieve, int autoUpdate, int autoDelete, boolean proxy)
    {
        ClassDescriptor cld = broker.getClassDescriptor(MovieImpl.class);
        CollectionDescriptor cod = (CollectionDescriptor) cld.getCollectionDescriptors().get(0);
        cod.setLazy(proxy);
        cod.setCascadeRetrieve(autoRetrieve);
        cod.setCascadingStore(autoUpdate);
        cod.setCascadingDelete(autoDelete);
    }


    //=======================================================================
    // Inner classes, persistence capable  test classes
    //=======================================================================
    public static class MovieManageableCollection implements ManageableCollection
    {
        private ArrayList list = new ArrayList();

        public void ojbAdd(Object anObject)
        {
            list.add(anObject);
        }

        public void ojbAddAll(ManageableCollection otherCollection)
        {
            Iterator it = otherCollection.ojbIterator();
            while(it.hasNext())
            {
                list.add(it.next());
            }
        }

        public Iterator ojbIterator()
        {
            return list.iterator();
        }

        public void afterStore(PersistenceBroker broker) throws PersistenceBrokerException
        {
        }

        public Iterator iterator()
        {
            return list.iterator();
        }

        public int size()
        {
            return list.size();
        }

        public boolean isEmpty()
        {
            return list.isEmpty();
        }

        public void clear()
        {
            list.clear();
        }

        public boolean add(Movie movie)
        {
            return list.add(movie);
        }

        public boolean remove(Movie movie)
        {
            return list.remove(movie);
        }

        public boolean contains(Movie movie)
        {
            return list.contains(movie);
        }

        public Movie get(int index)
        {
            return (Movie) list.get(index);
        }
    }

    public static class Actor
    {
        private Integer id;
        private Integer id2;
        private String name;
        private MovieManageableCollection movies;

        public Actor()
        {
        }

        public Actor(String name)
        {
            this.name = name;
        }

        public MovieManageableCollection getMovies()
        {
            return movies;
        }

        public void setMovies(MovieManageableCollection movies)
        {
            this.movies = movies;
        }

        public Integer getId()
        {
            return id;
        }

        public void setId(Integer id)
        {
            this.id = id;
        }

        public Integer getId2()
        {
            return id2;
        }

        public void setId2(Integer id2)
        {
            this.id2 = id2;
        }

        public String getName()
        {
            return name;
        }

        public void setName(String name)
        {
            this.name = name;
        }

        public String toString()
        {
            return ToStringBuilder.reflectionToString(this).toString();
        }
    }

    public static interface Movie
    {
        public Collection getActors();

        public void setActors(Collection actors);

        public Integer getIdInt2();
        public Integer getIdInt();

        public void setIdInt2(Integer id2Int);
        public void setIdInt(Integer idInt);

        public String getIdStr();

        public void setIdStr(String idStr);

        public String getTitle();

        public void setTitle(String title);

        public String getDescription();

        public void setDescription(String description);
    }

    public static class MovieImpl implements Movie
    {
        private Integer idInt;
        private Integer idInt2;
        private String idStr;
        private String title;
        private String description;
        private Collection actors;

        public MovieImpl()
        {
        }

        public MovieImpl(String idStr, String title, String description)
        {
            this.idStr = idStr;
            this.title = title;
            this.description = description;
        }

        public Collection getActors()
        {
            return actors;
        }

        public void setActors(Collection actors)
        {
            this.actors = actors;
        }

        public Integer getIdInt()
        {
            return idInt;
        }

        public void setIdInt(Integer idInt)
        {
            this.idInt = idInt;
        }

        public Integer getIdInt2()
        {
            return idInt2;
        }

        public void setIdInt2(Integer idInt2)
        {
            this.idInt2 = idInt2;
        }

        public String getIdStr()
        {
            return idStr;
        }

        public void setIdStr(String idStr)
        {
            this.idStr = idStr;
        }

        public String getTitle()
        {
            return title;
        }

        public void setTitle(String title)
        {
            this.title = title;
        }

        public String getDescription()
        {
            return description;
        }

        public void setDescription(String description)
        {
            this.description = description;
        }

        public String toString()
        {
            return ToStringBuilder.reflectionToString(this).toString();
        }
    }

    public static class Role
    {
        private Integer actorId;
        private Integer actorId2;
        private Integer movieIntId;
        private Integer movieIntId2;
        private String movieStrId;

        public Role()
        {
        }

        public Integer getActorId()
        {
            return actorId;
        }

        public void setActorId(Integer actorId)
        {
            this.actorId = actorId;
        }

        public Integer getMovieIntId()
        {
            return movieIntId;
        }

        public Integer getMovieIntId2()
        {
            return movieIntId2;
        }

        public void setMovieIntId2(Integer movieIntId2)
        {
            this.movieIntId2 = movieIntId2;
        }

        public Integer getActorId2()
        {
            return actorId2;
        }

        public void setActorId2(Integer actorId2)
        {
            this.actorId2 = actorId2;
        }

        public void setMovieIntId(Integer movieIntId)
        {
            this.movieIntId = movieIntId;
        }

        public String getMovieStrId()
        {
            return movieStrId;
        }

        public void setMovieStrId(String movieStrId)
        {
            this.movieStrId = movieStrId;
        }

        public String toString()
        {
            return ToStringBuilder.reflectionToString(this).toString();
        }
    }
}
