package org.apache.ojb.performance;

/* Copyright 2002-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.OutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * The OJB stress/performance test - a simple performance test framework for
 * testing multi-threaded environments.
 *
 * <p>
 * <b>You have two possibilities to run this test:</b>
 * </p>
 * <p>
 * - use the build script and call
 * <br/>
 * <code>ant perf-test</code>
 * </p>
 * <p>
 * - or perform the test class by yourself
 * <br/>
 * <code>
 * java -classpath CLASSPATH org.apache.ojb.performance.PerfMain
 * </code>
 * <br/>
 * <code>
 * [comma separated list of PerfTest implementation classes, no blanks!]
 * </code>
 * <br/>
 * <code>
 * [number of test loops]
 * </code>
 * <br/>
 * <code>
 * [number of threads]
 * </code>
 * <br/>
 * <code>
 * [number of insert/fetch/delete loops per thread]
 * </code>
 * <br/>
 * <code>
 * [boolean - run in stress mode if set true, run in performance mode if set false, default false]
 * </code>
 * </p>
 * <p>
 * For example:
 * </p>
 * <code>java -classpath CLASSPATH my.MyPerfTest,myMyPerfTest2 3 10 2000 false</code>
 *
 * @author <a href="mailto:armin@codeAuLait.de">Armin Waibel</a>
 * @version $Id: PerfMain.java,v 1.9 2004/06/14 12:47:29 arminw Exp $
 */
public class PerfMain
{
    protected static final String EOL = System.getProperty("line.separator");

    private static final String PREFIX_LOG = "[" + PerfMain.class.getName() + "] ";
    /**
     * iterations per thread
     */
    private static int iterationsPerThread = 500;
    /**
     * number of concurrent threads
     */
    private static int concurrentThreads = 10;
    /**
     * if false we use performance optimized delete/insert method
     */
    private static boolean useStressMode = false;
    /**
     * number of test loops
     */
    private static int testLoops = 1;

    private HashMap resultMap;
    private Map exceptionMap;


    public static void main(String[] args)
    {
        PerfMain main = new PerfMain();
        try
        {
            long peroid = System.currentTimeMillis();
            // start the test
            main.startPerfTest(args);
            // print the test results
            main.printResult(System.out);
            peroid = System.currentTimeMillis() - peroid;
            System.out.println();
            System.out.println("PerfTest takes " + peroid / 1000 + " [sec]");
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        System.exit(0);
    }

    public PerfMain()
    {
        this.resultMap = new HashMap();
        this.exceptionMap = new Hashtable();
    }

    /**
     * Call this to begin the performance test.
     */
    public void startPerfTest(String[] args) throws Exception
    {
        ArrayList testList = null;
        try
        {
            // comma separated list of the PerfTest implementation classes
            if (args.length > 0)
            {
                StringTokenizer tok = new StringTokenizer(args[0], ",");
                testList = new ArrayList();
                while (tok.hasMoreTokens())
                {
                    testList.add(tok.nextToken().trim());
                }
            }
            else
            {
                throw new IllegalArgumentException("No test handles found!");
            }
            // number of test loops
            if (args.length > 1)
            {
                testLoops = args.length > 1 ? Integer.parseInt(args[1]) : 1;
            }
            // number of threads
            if (args.length > 2)
            {
                concurrentThreads = Integer.parseInt(args[2]);
            }
            // number of insert/fetch/delete loops per thread
            if (args.length > 3)
            {
                iterationsPerThread = Integer.parseInt(args[3]);
            }
            // run in stress mode
            if (args.length > 4)
            {
                useStressMode = Boolean.valueOf(args[4]).booleanValue();
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
            System.err.println();
            System.err.println("Usage of PerfMain:" +
                    "java -classpath CLASSPATH org.apache.ojb.performance.PerfMain" +
                    " [comma separated list of PerfTest implementation classes]" +
                    " [number of test loops]" +
                    " [number of threads]" +
                    " [number of insert/fetch/delete loops per thread]" +
                    " [boolean - run in stress mode]");
            System.err.println();
            System.err.println("Example: java -classpath" +
                    " CLASSPATH org.apache.ojb.performance.PerfMain org.MyPerfTest 3 10 500 false");
        }

        System.out.println("####################################################" +
                EOL + "# Start OJB performance-test framework - do " + testLoops + " loop #" +
                EOL + "####################################################" + EOL);

        Iterator it = testList.iterator();
        while (it.hasNext())
        {
            String perfTest = (String) it.next();
            PerfTest test = (PerfTest) Class.forName(perfTest).newInstance();
            test.registerPerfMain(this);
            Runtime rt = Runtime.getRuntime();
            long freeMem = 0;
            for (int i = 0; i < testLoops; i++)
            {
                System.out.println("perform test loop " + (i + 1));
                freeMem = rt.freeMemory();
                test.performTest();
                freeMem = (freeMem - rt.freeMemory()) / 1024;
                System.out.println("allocated memory after test: " + freeMem + " kb");
                rt.gc();
            }
        }
    }

    public void printResult(OutputStream out)
    {
        PrintStream print = new PrintStream(out);
        print.println();

        if (!getExceptionMap().isEmpty())
        {
            StringBuffer buf = new StringBuffer();
            buf.append(EOL + "Failures occured, test not valid:" + EOL);
            Iterator it = getExceptionMap().keySet().iterator();
            while (it.hasNext())
            {
                String causer = (String) it.next();
                buf.append("Failure cause by " + causer + ", exception was " + exceptionMap.get(causer) + EOL);
            }
            print.println(buf.toString());
        }

        print.println(buildTestSummary(resultMap.values()));
    }

    private String buildTestSummary(Collection results)
    {
        int column = 8;
        StringBuffer buf = new StringBuffer();
        // table header
        buf.append(EOL);
        for (int i = 0; i < 7; i++)
        {
            buf.append(getFilledUpToLength("", column, "="));
        }
        buf.append(EOL);
        buf.append(getFilledUpToLength("", column, " "));
        buf.append(getFilledUpToLength("OJB PERFORMANCE TEST SUMMARY", column, " "));
        buf.append(EOL);

        buf.append(PerfMain.getConcurrentThreads() + " concurrent threads, handle " +
                PerfMain.getIterationsPerThread() + " objects per thread");
        buf.append(EOL);
        buf.append(getFilledUpToLength("", column, " "));
        buf.append("- " + (isUseStressMode() == false ? "performance mode" : "stress mode -"));
        buf.append(EOL);
        for (int i = 0; i < 7; i++)
        {
            buf.append(getFilledUpToLength("", column, "="));
        }
        buf.append(EOL);
        buf.append(getFilledUpToLength("API", column, " "));
        buf.append(getFilledUpToLength("Period", column, " "));
        buf.append(getFilledUpToLength("Total", column, " "));
        buf.append(getFilledUpToLength("Insert", column, " "));
        buf.append(getFilledUpToLength("Fetch", column, " "));
        buf.append(getFilledUpToLength("Update", column, " "));
        buf.append(getFilledUpToLength("Delete", column, " "));
        buf.append(EOL);
        buf.append(getFilledUpToLength("", column, " "));
        buf.append(getFilledUpToLength("[sec]", column, " "));
        buf.append(getFilledUpToLength("[sec]", column, " "));
        buf.append(getFilledUpToLength("[msec]", column, " "));
        buf.append(getFilledUpToLength("[msec]", column, " "));
        buf.append(getFilledUpToLength("[msec]", column, " "));
        buf.append(getFilledUpToLength("[msec]", column, " "));
        buf.append(EOL);
        for (int i = 0; i < 7; i++)
        {
            buf.append(getFilledUpToLength("", column, "-"));
        }

        // fill table
        Iterator it = results.iterator();
        while (it.hasNext())
        {
            buf.append(EOL);
            PerfResult res = (PerfResult) it.next();
            buf.append(getFilledUpToLength(res.getTestName(), column, " "));
            buf.append(getFilledUpToLength(new Double((double) (((double) res.getTestPeriod()) / 1000)).toString(), column, " "));
            buf.append(getFilledUpToLength(new Double((double) (((double) res.getTotalTime()) / 1000)).toString(), column, " "));
            buf.append(getFilledUpToLength(new Long(res.getInsertPeriod()).toString(), column, " "));
            buf.append(getFilledUpToLength(new Long(res.getFetchPeriod()).toString(), column, " "));
            buf.append(getFilledUpToLength(new Long(res.getUpdatePeriod()).toString(), column, " "));
            buf.append(getFilledUpToLength(new Long(res.getDeletePeriod()).toString(), column, " "));
        }
        buf.append(EOL);
        for (int i = 0; i < 7; i++)
        {
            buf.append(getFilledUpToLength("", column, "="));
        }
        return buf.toString();
    }

    private String getFilledUpToLength(String target, int length, String fillCharacter)
    {
        if (target.length() > length) return target;

        int count = length - target.length();
        String blanks = "";
        for (int i = 0; i < count; i++)
        {
            blanks = blanks + fillCharacter;
        }
        return blanks + target;
    }

    private String buildObjectBeforeAfterList(List consistentList)
    {
        StringBuffer buf = new StringBuffer();
        buf.append("Test failed, inconsistent object count:");
        buf.append(EOL);
        for (int i = 0; i < consistentList.size(); i++)
        {
            ConsistentEntry entry = (ConsistentEntry) consistentList.get(i);
            buf.append("passed: " + entry.isPassed() + " - objects before test/ after test   " +
                    entry.getObjectsBefore() + "/" + entry.getObjectsAfter() + EOL);
        }
        return buf.toString();
    }

    /**
     * resultArr[0] startTime/test length
     * resultArr[1] inserting times
     * resultArr[2] fetching times
     * resultArr[3] updating times
     * resultArr[4] deleting times
     */
    public void addPeriodResult(String testName, long[] resultArr)
    {
        PerfResult result = (PerfResult) resultMap.get(testName);
        if (result == null)
        {
            result = new PerfResult();
            result.setTestName(testName);
            result.setStressMode(isUseStressMode());
            result.setIterationsPerThread(getIterationsPerThread());
            result.setNumberOfThreads(getConcurrentThreads());
            result.setTestLoops(getTestLoops());
            resultMap.put(testName, result);

        }
        result.addTestPeriod(resultArr[0]);
        result.addInsertPeriod(resultArr[1]);
        result.addFetchPeriod(resultArr[2]);
        result.addUpdatePeriod(resultArr[3]);
        result.addDeletePeriod(resultArr[4]);
    }

    public void addConsistentResult(String testName, int objectsBefore, int objectsAfter)
    {
        ConsistentEntry ce = new ConsistentEntry(objectsBefore, objectsAfter);
        PerfResult result = (PerfResult) resultMap.get(testName);
        result.addConsistentEntry(ce);
    }

    public synchronized void registerException(String causer, Exception e)
    {
        exceptionMap.put(causer, e);
    }

    public Map getExceptionMap()
    {
        return exceptionMap;
    }

    public Collection getResultList()
    {
        return resultMap.values();
    }

    public static int getIterationsPerThread()
    {
        return iterationsPerThread;
    }

    public static int getConcurrentThreads()
    {
        return concurrentThreads;
    }

    public static boolean isUseStressMode()
    {
        return useStressMode;
    }

    public static int getTestLoops()
    {
        return testLoops;
    }

    //================================================================
    // inner class
    //================================================================
    static class PerfResult
    {
        private String testName;
        private long testPeriod;
        private boolean stressMode;

        private int testLoops;
        private int numberOfThreads;
        private int iterationsPerThread;

        private long insertPeriod;
        private long fetchPeriod;
        private long updatePeriod;
        private long deletePeriod;

        private boolean valid;

        private List consistentList;

        public PerfResult()
        {
            setValid(true);
            this.consistentList = new ArrayList();
        }

        public String toString()
        {
            StringBuffer buf = new StringBuffer();
            buf.append(EOL + "[" + this.getClass().getName());
            buf.append(EOL + "testName=" + testName);
            buf.append(EOL + "testPeriod=" + testPeriod);
            buf.append(EOL + "testLoops=" + testLoops);
            buf.append(EOL + "numberOfThreads=" + numberOfThreads);
            buf.append(EOL + "iterationsPerThread=" + iterationsPerThread);
            buf.append(EOL + "isValid=" + isValid());
            buf.append(EOL + "insertPeriod=" + getInsertPeriod());
            buf.append(EOL + "fetchPeriod=" + getFetchPeriod());
            buf.append(EOL + "deletePeriod=" + getDeletePeriod());
            buf.append(EOL + "consistentList: " + consistentList);
            buf.append("]");
            return buf.toString();
        }

        public void addConsistentEntry(ConsistentEntry entry)
        {
            this.consistentList.add(entry);
            valid = valid && entry.isPassed();
        }

        public boolean isStressMode()
        {
            return stressMode;
        }

        public void setStressMode(boolean stressMode)
        {
            this.stressMode = stressMode;
        }

        public boolean isValid()
        {
            return valid;
        }

        public void setValid(boolean valid)
        {
            this.valid = valid;
        }

        public String getTestName()
        {
            return testName;
        }

        public void setTestName(String testName)
        {
            this.testName = testName;
        }

        public long getTestPeriod()
        {
            return testPeriod / getTestLoops();
        }

        public void addTestPeriod(long aTestPeriod)
        {
            this.testPeriod += aTestPeriod;
        }

        public int getTestLoops()
        {
            return testLoops;
        }

        public void setTestLoops(int testLoops)
        {
            this.testLoops = testLoops;
        }

        public int getNumberOfThreads()
        {
            return numberOfThreads;
        }

        public void setNumberOfThreads(int numberOfThreads)
        {
            this.numberOfThreads = numberOfThreads;
        }

        public int getIterationsPerThread()
        {
            return iterationsPerThread;
        }

        public void setIterationsPerThread(int numberOfObjects)
        {
            this.iterationsPerThread = numberOfObjects;
        }

        public long getTotalTime()
        {
            return ((insertPeriod + fetchPeriod + updatePeriod + deletePeriod) / getTestLoops()) / getNumberOfThreads();
        }

        public long getInsertPeriod()
        {
            return (insertPeriod / getTestLoops()) / getNumberOfThreads();
        }

        public void addInsertPeriod(long anInsertPeriod)
        {
            this.insertPeriod += anInsertPeriod;
        }

        public long getFetchPeriod()
        {
            return (fetchPeriod / getTestLoops()) / getNumberOfThreads();
        }

        public void addFetchPeriod(long aFetchPeriod)
        {
            this.fetchPeriod += aFetchPeriod;
        }

        public long getUpdatePeriod()
        {
            return (updatePeriod / getTestLoops()) / getNumberOfThreads();
        }

        public void addUpdatePeriod(long aUpdatePeriod)
        {
            this.updatePeriod += aUpdatePeriod;
        }

        public long getDeletePeriod()
        {
            return (deletePeriod / getTestLoops()) / getNumberOfThreads();
        }

        public void addDeletePeriod(long aDeletePeriod)
        {
            this.deletePeriod += aDeletePeriod;
        }
    }

    //================================================================
    // inner class
    //================================================================
    static class ConsistentEntry
    {
        private int objectsBefore;
        private int objectsAfter;
        private boolean passed;

        public ConsistentEntry(int objectsBefore, int objectsAfter)
        {
            this.objectsBefore = objectsBefore;
            this.objectsAfter = objectsAfter;
        }

        public int getObjectsBefore()
        {
            return objectsBefore;
        }

        public int getObjectsAfter()
        {
            return objectsAfter;
        }

        public boolean isPassed()
        {
            return objectsBefore == objectsAfter;
        }

        public String toString()
        {
            StringBuffer buf = new StringBuffer();
            buf.append("[" + this.getClass().getName()).
                    append(": objectsBefore=" + getObjectsBefore()).
                    append(" objectsAfter=" + objectsAfter).
                    append(" isPassed=" + isPassed());
            return buf.toString();
        }
    }
}
