package org.apache.ojb.performance;

/* Copyright 2002-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Derivate this class to implement a test instance for the performance test.
 *
 * @author <a href="mailto:armin@codeAuLait.de">Armin Waibel</a>.
 * @version $Id: PerfTest.java,v 1.9 2004/04/05 12:16:22 tomdz Exp $
 */
public abstract class PerfTest
{
    private final String PREFIX_LOG = "[" + this.getClass().getName() + "] ";

    /**
     * testTimes[0] startTime/test length
     * testTimes[1] inserting times
     * testTimes[2] fetching times
     * testTimes[3] updating times
     * testTimes[4] deleting times
     */
    private long[] testTimes;
    private int threadCount;
    private ThreadGroup threadGroup;
    private PerfMain perfMain;
    private long perfTestId;
    private boolean checked;
    /**
     * The threads that are executing.
     */
    private Thread threads[] = null;

    public PerfTest()
    {
        threadGroup = new ThreadGroup(testName() + "_Group");
        perfTestId = System.currentTimeMillis();
        checked = false;
    }

    /**
     * Returns the name of the test
     */
    public abstract String testName();

    /**
     * Returns a new instance of a {@link PerfHandle} implementation.
     */
    public abstract PerfHandle newPerfHandle(PerfTest test);

    /**
     * Returns the count of all found {@link PerfArticle}
     * in database. Do not use the cache to get the count!
     * This method is not involved in the performance test
     * methods, thus it's not mandatory to use the api-methods
     * for implementation.
     */
    public abstract int articleCount();


    private void checkApi() throws Exception
    {
        String name = testName() + "_Pre_Test_Object";
        PerfHandle handle = newPerfHandle(this);
        PerfArticle article = handle.getPreparedPerfArticle(name);
        PerfArticle[] arr = new PerfArticle[]{article};
        handle.init();
        handle.insertNewArticles(arr);
        handle.readArticlesByCursor(name);
        handle.updateArticles(arr);
        handle.deleteArticles(arr);
        handle.tearDown();
        checked = true;
    }

    /**
     * Interrupt the running threads.
     */
    protected void interruptThreads()
    {
        if (threads != null)
        {
            for (int i = 0; i < threads.length; i++)
            {
                threads[i].interrupt();
            }
        }
        System.err.println("## Test failed! ##");
        System.err.println("## Test failed! ##");
    }

    /**
     * Run the threads.
     */
    protected void runTestHandles(final PerfHandle[] runnables)
    {
        if (runnables == null)
        {
            throw new IllegalArgumentException("runnables is null");
        }
        threads = new Thread[runnables.length];
        for (int i = 0; i < threads.length; i++)
        {
            threads[i] = new Thread(runnables[i]);
        }
        for (int i = 0; i < threads.length; i++)
        {
            threads[i].start();
        }
        try
        {
            for (int i = 0; i < threads.length; i++)
            {
                threads[i].join();
            }
        }
        catch (InterruptedException ignore)
        {
            System.out.println(PREFIX_LOG + "Thread join interrupted.");
        }
        threads = null;
    }

    public void performTest()
    {
        try
        {
            if (!checked)
            {
                checkApi();
                System.out.println(PerfMain.EOL + "# Start PerfTest: " + testName() + " #");
            }

            int objectCount = 0;
            int objectCountAfter = 0;

            testTimes = new long[5];
            threadCount = PerfMain.getConcurrentThreads();

            objectCount = articleCount();

            // now we start the test threads
            PerfHandle[] perfHandles = new PerfHandle[PerfMain.getConcurrentThreads()];
            for (int i = 0; i < PerfMain.getConcurrentThreads(); i++)
            {
                PerfHandle ph = newPerfHandle(this);
                perfHandles[i] = ph;
            }
            testTimes[0] = System.currentTimeMillis();
            runTestHandles(perfHandles);
            testTimes[0] = System.currentTimeMillis() - testTimes[0];
            // end of test threads

            objectCountAfter = articleCount();
            perfMain.addPeriodResult(testName(), testTimes);
            perfMain.addConsistentResult(testName(), objectCount, objectCountAfter);
        }
        catch (Exception e)
        {
            e.printStackTrace();
            perfMain.registerException(PREFIX_LOG, e);
        }
    }

    public void registerException(String causer, Exception e)
    {
        perfMain.registerException(causer, e);
    }

    public synchronized void addTime(int position, long time)
    {
        testTimes[position] = testTimes[position] + time;
    }

    public void registerPerfMain(PerfMain aPerfMain)
    {
        this.perfMain = aPerfMain;
    }

    public ThreadGroup getThreadGroup()
    {
        return threadGroup;
    }

    public long getPerfTestId()
    {
        return perfTestId;
    }
}
